@@ -1,13 +1,13 @@
 // Copyright (C) 2016 Kurt Pattyn <pattyn.kurt@gmail.com>.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 #include <QString>
 #include <QtTest>
 #include <QNetworkProxy>
 #include <QTcpSocket>
 #include <QTcpServer>
 #include <QtCore/QScopedPointer>
-#include <QtNetwork/qsslpresharedkeyauthenticator.h>
 #ifndef QT_NO_SSL
+#include <QtNetwork/qsslpresharedkeyauthenticator.h>
 #include <QtNetwork/qsslcipher.h>
 #include <QtNetwork/qsslkey.h>
 #include <QtNetwork/qsslsocket.h>
@@ -44,6 +44,7 @@ public:
     QByteArray m_psk;
 
 public slots:
+#if QT_CONFIG(ssl)
     void providePsk(QSslPreSharedKeyAuthenticator *authenticator)
     {
         QVERIFY(authenticator);
@@ -65,6 +66,7 @@ public slots:
             QCOMPARE(authenticator->preSharedKey(), m_psk);
         }
     }
+#endif
 };
 
 class tst_QWebSocketServer : public QObject
@@ -84,12 +86,15 @@ private Q_SLOTS:
     void tst_connectivity();
     void tst_protocols_data();
     void tst_protocols();
+#if QT_CONFIG(ssl)
     void tst_preSharedKey();
+#endif
     void tst_maxPendingConnections();
     void tst_serverDestroyedWhileSocketConnected();
     void tst_scheme(); // qtbug-55927
     void tst_handleConnection();
     void tst_handshakeTimeout(); // qtbug-63312, qtbug-57026
+    void gentleClose();
     void multipleFrames();
 
 private:
@@ -256,7 +261,7 @@ void tst_QWebSocketServer::tst_settersAndGetters()
 
     server.setHandshakeTimeout(64);
     QCOMPARE(server.handshakeTimeoutMS(), 64);
-#if QT_HAS_INCLUDE(<chrono>)
+#if __has_include(<chrono>)
     auto expected = std::chrono::milliseconds(64);
     QCOMPARE(server.handshakeTimeout(), expected);
 
@@ -409,6 +414,7 @@ void tst_QWebSocketServer::tst_protocols()
     QCOMPARE(serverSocket->handshakeOptions().subprotocols(), clientProtocols + headerProtocols);
 }
 
+#if QT_CONFIG(ssl)
 void tst_QWebSocketServer::tst_preSharedKey()
 {
     if (m_shouldSkipUnsupportedIpv6Test)
@@ -491,6 +497,7 @@ void tst_QWebSocketServer::tst_preSharedKey()
     QCOMPARE(sslErrorsSpy.size(), 0);
     QCOMPARE(serverErrorSpy.size(), 0);
 }
+#endif
 
 void tst_QWebSocketServer::tst_maxPendingConnections()
 {
@@ -811,7 +818,7 @@ void tst_QWebSocketServer::tst_handshakeTimeout()
         QCOMPARE(secureServerConnectionSpy.size(), 0);
 
         QWebSocket secureSocket;
-        connect(&secureSocket, QOverload<QAbstractSocket::SocketError>::of(&QWebSocket::error),
+        connect(&secureSocket, &QWebSocket::errorOccurred,
                 [](QAbstractSocket::SocketError error) {
                     // This shouldn't print but it's useful for debugging when/if it does.
                     qDebug() << "Error occurred in the client:" << error;
@@ -855,6 +862,37 @@ void tst_QWebSocketServer::tst_handshakeTimeout()
     }
 }
 
+void tst_QWebSocketServer::gentleClose()
+{
+    QWebSocketServer server("gentleClose", QWebSocketServer::NonSecureMode);
+    QSignalSpy serverConnectionSpy(&server, &QWebSocketServer::newConnection);
+    QVERIFY(server.listen());
+
+    QWebSocket client;
+    QSignalSpy connectedSpy(&client, &QWebSocket::connected);
+    QSignalSpy clientDisconnectedSpy(&client, &QWebSocket::disconnected);
+    QSignalSpy clientErrorSpy(&client, &QWebSocket::errorOccurred);
+    client.open(server.serverUrl());
+
+    QVERIFY(serverConnectionSpy.wait());
+    QVERIFY(connectedSpy.wait());
+    QWebSocket *serverSocket = server.nextPendingConnection();
+    QSignalSpy disconnectedSpy(serverSocket, &QWebSocket::disconnected);
+    QSignalSpy serverErrorSpy(serverSocket, &QWebSocket::errorOccurred);
+
+    client.close(QWebSocketProtocol::CloseCodeNormal);
+
+    QVERIFY(disconnectedSpy.wait());
+    QVERIFY(clientDisconnectedSpy.wait());
+
+    QVERIFY(client.closeCode() != QWebSocketProtocol::CloseCodeGoingAway);
+    QVERIFY(serverSocket->closeCode() != QWebSocketProtocol::CloseCodeGoingAway);
+    QVERIFY(client.error() != QAbstractSocket::RemoteHostClosedError);
+    QVERIFY(serverSocket->error() != QAbstractSocket::RemoteHostClosedError);
+    QCOMPARE(clientErrorSpy.count(), 0);
+    QCOMPARE(serverErrorSpy.count(), 0);
+}
+
 void tst_QWebSocketServer::multipleFrames()
 {
     QWebSocketServer server(QString(), QWebSocketServer::NonSecureMode);
